/*
 * $Id: normalize.c,v 1.3 2007/01/10 12:58:43 t-ishii Exp $
 *
 * Copyright (c) 2007	SRA Tohoku, Inc.
 * Copyright (c) 2007	SRA OSS, Inc. Japan
 *
 * Result of Open Source Software Development Activities of
 * Information-technology Promotion Agency, Japan.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above
 * copyright notice and this paragraph and the following two
 * paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS
 * IS" BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include "postgres.h"

#include "funcapi.h"
#include "utils/builtins.h"
#include "mb/pg_wchar.h"

#include "normalize.h"

PG_MODULE_MAGIC;

PG_FUNCTION_INFO_V1(normalize);

extern Datum normalize(PG_FUNCTION_ARGS);

/* ----------
 * normalize:
 * returns normalized text.
 * assumes database encoding is EUC_JP. otherwise raise an error.
 * CAUTION: encoding of this file MUST be EUC-JP!
 * this function does following:
 * - remove \r or \n
 * - convert JIS X 0208 0-9, alphabet to corresponding ASCII characters
 * - convert JIS X 0201 kana to JIS X 0208 kana
 * - convert ASCII capital characters to lower characters
 *
 * C FUNCTION definition
 * normalize(text) returns text
 *
 * ----------
 */

#define PG_TEXT_GET_STR(textp_) \
	DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp_)))
#define PG_STR_GET_TEXT(str_) \
	DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(str_)))
#define min(a, b) (a < b? a: b)

#define IS_NUMERIC(c) (c) >= 0xa3b0 && (c) <= 0xa3b9	/* JIS X 0208 numeric? */
#define TO_ASCII_NUMERIC(c) ((c) - 0xa3b0 + 0x30)		/* JIS X 0208 numeric to ASCII */
#define IS_ALPHA_UPPER(c) (c) >= 0xa3c1 && (c) <= 0xa3da		/* JIS X 0208 upper case alphabet? */
#define TO_ALPHA_UPPER_TO_LOWER(c) ((c) - 0xa3c1 + 0x61)		/* JIS X 0208 upper case alphabet to ASCII lower case */
#define IS_ALPHA_LOWER(c) (c) >= 0xa3e1 && (c) <= 0xa3fa		/* JIS X 0208 lower case alphabet? */
#define TO_ALPHA_LOWER_TO_LOWER(c) ((c) - 0xa3e1 + 0x61)		/* JIS X 0208 lower case alphabet to ASCII lower case */
#define IS_KANA(c) (c) >= 0x8ea6 && (c) <= 0x8edf		/* JIS X 0201 kana? */

static unsigned char *to_kana(unsigned char *kana, int *len);

Datum
normalize(PG_FUNCTION_ARGS)
{
	unsigned char *src;
	unsigned char *dest;
	unsigned char *d;
	text *t;
	text *result;
	unsigned short w;
	int len;

	if (strcmp(GetDatabaseEncodingName(), "EUC_JP"))
	{
		elog(ERROR, "normalize() only works with EUC_JP encoding");
	}

	t = PG_GETARG_TEXT_P(0);
	src = PG_TEXT_GET_STR(t);

	/*
	 * we assume that the converted string is no longer than
	 * the source string
	 */
	d = dest = palloc(VARSIZE(t) - VARHDRSZ + 1);
	*d = '\0';

	while (*src)
	{
		/* get multibyte length */
		len = pg_mblen(src);

		/* omit line feed etc. */
		if (*src == '\r' || *src == '\n')
		{
			while (*src == '\r' || *src == '\n')
			{
				/* insert space if there's a empty line */
				if (*src == '\r' && *(src+1) == '\r')	/* Mac style */
				{
					*d++ = ' ';
					src ++;
				}
				else if (*src == '\n' && *(src+1) == '\n')	/* Unix style */
				{
					*d++ = ' ';
					src ++;
				}
				else if (*src == '\r' && *(src+1) == '\n' &&
						 *(src+2) == '\r' && *(src+3) == '\n')	/* DOS style */
				{
					*d++ = ' ';
					src += 3;
				}

				src++;
			}
			
			/* insert a space if the last char in the previous
			 * line or beginning of this line are ASCII
			 */
			if ((d != dest && *(d-1) && *(d-1) <= 0x7f) || (*src && *src <= 0x7f))
				*d++ = ' ';

			continue;
		}

		else if (len == 1)
		{
			if (isalpha(*src))	/* ASCII alphabet? */
			{
				*d++ = tolower(*src++);		/* convert to lower case */
			}
			else
				*d++ = *src++;	/* not alphabet. just copy it */
			continue;
		}

		/*
		 * non ASCII case
		 */
		else
		{
			/* to make sure that multibyte string is not a broken one */
			if (*(src+1) != '\0')
			{
				int kanalen;
				unsigned char *p;

				/* JIS X 0201 kana conversion */
				if ((p = to_kana(src, &kanalen)))
				{
					/* success in conversion */
					*d++ = *p++;
					*d++ = *p++;
					src += kanalen;
					continue;
				}

				w = *src << 8 | *(src+1);

				if (IS_NUMERIC(w))
					w = TO_ASCII_NUMERIC(w);
				else if (IS_ALPHA_UPPER(w))
					w = TO_ALPHA_UPPER_TO_LOWER(w);
				else if (IS_ALPHA_LOWER(w))
					w = TO_ALPHA_LOWER_TO_LOWER(w);
				else
				{
					while (*src && len--)
						*d++ = *src++;
					continue;
				}

				/* convert to ASCII */
				*d++ = w & 0xff;
				src += 2;
				continue;
			}
			*d++ = *src++;		/* just copy broken 1 byte... */
		}
	}
	*d = '\0';
	result = PG_STR_GET_TEXT(dest);
	pfree(dest);
	PG_RETURN_TEXT_P(result);
}

/*
 * convert a JIS X 0201 kana to JIS X 0208 kana
 * kana is JIS X 0201 kana
 * consumed kana byte length is stored into *len
 */
	static unsigned char *to_kana(unsigned char *kana, int *len)
{
	static unsigned char *JIS0201kana = "ÎĎŎƎǎȎɎʎˎ͎̎ΎώЎюҎӎԎՎ֎׎؎َڎێ܎";
	static unsigned char *JIS0208kana = "åĥƥȥʥ˥̥ͥΥϥҥեإۥޥߥ";
	static unsigned char *JIS0201kanadaku = "ގގގގގގގގގގގގގÎގĎގʎގʎߎˎގˎߎ̎ގ̎ߎ͎ގ͎ߎΎގΎߎ";
	static unsigned char *JIS0208kanadaku = "¥ťǥɥХѥӥԥ֥ץ٥ڥܥݥ";

	unsigned char *p, *k;

	*len = 0;
	k = kana;

	if (*k == '\0' || *(k+1) == '\0')
		return NULL;

	if (*(k+2) && *(k+3))
	{
		p = JIS0201kanadaku;

		while (*p)
		{
			if (*k == *p && *(k+1) == *(p+1) &&
				*(k+2) == *(p+2) && *(k+3) == *(p+3))
			{
				/* found */
				*len = 4;
				return JIS0208kanadaku + (p - JIS0201kanadaku)/2;
			}
			p += 4;
		}
	}

	p = JIS0201kana;

	while (*p)
	{
		if (*k == *p && *(k+1) == *(p+1))
		{
			/* found */
			*len = 2;
			return JIS0208kana + (p - JIS0201kana);
		}
		p += 2;
	}
	return NULL;
}
